#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/wait.h>
#include <fcntl.h>
#include "myftp.h"


//TODO: Add comment about getting server code from assignment 8
//TODO: Final pass on print statements, which to keep, remove, or make debugging


int main(int argc, char* argv[]){
	fprintf(stderr, "Server starting\n");
	int listenFD;
	setUpServer(&listenFD);

	while(1){
		int connectfd;
		if(serverFork(listenFD, &connectfd) != 0) continue;

	    int datafd = -1;
		while(1){
			char command[CMDLEN];
			receiveCommand(connectfd, command);
	        if(DEBUG) fprintf(stderr, "Received Command:\n-----%c-----\nArgument:\n-----%s-----\n", command[0], command+1);

	        int ret;
	        switch(command[0]){
	        	case 'D': ret = DCMD(&datafd, connectfd); break;
	        	case 'C': ret = CCMD(command+1, connectfd); break;
	        	case 'L': ret = LCMD(&datafd, connectfd); break;
	        	case 'G': ret = GCMD(command+1, &datafd, connectfd); break;
	        	case 'P': ret = PCMD(command+1, &datafd, connectfd); break;
	        	case 'Q': sendString(connectfd, "A\n", 2); exit(0);
	        	default: error("Invalid command", 1, command);
	        }
	        if(ret == 0) continue;

	        sprintf(command, "E%s\n", strerror(ret));
	        sendString(connectfd, command, strlen(command));
	    }
	    fprintf(stderr, "Closing connection\n");
	}
	return 0;
}


void setUpServer(int* listenFD){
	struct sockaddr_in servAddr;
	setUpSocket(listenFD, &servAddr);

	if(bind(*listenFD, (struct sockaddr*)&servAddr, sizeof(servAddr)) != 0)
		error(strerror(errno), 1, "Se1");
	if(listen(*listenFD, 1) == -1) error(strerror(errno), 1, "Se3");
}


pid_t serverFork(int listenFD, int* connectfd){
	int length = sizeof(struct sockaddr_in);
	struct sockaddr_in clientAddr;
	if(((*connectfd = accept(listenFD, (struct sockaddr*)&clientAddr, &length)) == -1))
        error(strerror(errno), 1, "Se2");

    char hostName[NI_MAXHOST];
    int hostEntry = getnameinfo((struct sockaddr*)&clientAddr, sizeof(clientAddr), 
                                hostName, sizeof(hostName), NULL, 0, NI_NUMERICSERV);
    printf("\"%s\" Connected\n", hostName);

	pid_t output;
    if((output = fork()) != 0){
    	close(*connectfd);
    	while(waitpid(0, NULL, WNOHANG));
    }
    return output;
}


void receiveCommand(int FD, char* string){
	if(DEBUG) fprintf(stderr, "Start Read\n");
	
	int len = 0;
	while(1){
		int chars = read(FD, string+len, CMDLEN-len);
		if(chars == 0) {
			if(len == 0) error("Client Didn't Respond", 1, "Se13");
			return;
		}
		if(chars == -1) error(strerror(errno), 1, "Se4");
		
		len += chars;
		if(string[len-1] == '\n'){
			string[len-1] = '\0';
			return;
		}
	}
}


int CCMD(char* argument, int connectFD){
	if(chdir(argument) == -1) return errno;
	sendString(connectFD, "A\n", 2);
	return 0;
}


int DCMD(int* dataFD, int connectFD){
	int tempFD;
	if((tempFD = socket(AF_INET, SOCK_STREAM, 0)) == -1)
        error(strerror(errno), 1, "Se5");

    struct sockaddr_in servAddr = {0};
    servAddr.sin_family = AF_INET;
    servAddr.sin_addr.s_addr = htons(INADDR_ANY);

    if((bind(tempFD, (struct sockaddr*)&servAddr, sizeof(struct sockaddr_in))) != 0)
		error(strerror(errno), 1, "Se6");

	int addrLen = sizeof(struct sockaddr_in);
	if(getsockname(tempFD, (struct sockaddr*)&servAddr, &addrLen) == -1)
		error(strerror(errno), 1, "Se7");

	if(listen(tempFD, 1) == -1) error(strerror(errno), 1, "Se9");

	char string[CMDLEN];
	sprintf(string, "A%i\n", ntohs(servAddr.sin_port));
	sendString(connectFD, string, strlen(string));

	if(((*dataFD = accept(tempFD, (struct sockaddr*)&servAddr, &addrLen)) == -1))
        error(strerror(errno), 1, "Se8");

	return 0;
}


int LCMD(int* dataFD, int connectFD){
	if(*dataFD == -1) error("L command issued without a data connection", 1, "Se10");

	if(fork() == 0){
		if(close(1) == -1) error(strerror(errno), 1, "Se21");
		dup(*dataFD);
		execlp("ls", "ls", "-l", NULL);
		error(strerror(errno), 1, "Se11");
	}
	wait(NULL);

	if(close(*dataFD) == -1) error(strerror(errno), 1, "Se22");
	*dataFD = -1;
	sendString(connectFD, "A\n", 2);
	return 0;
}


int GCMD(char* argument, int* dataFD, int connectFD){
	if(*dataFD == -1) error("G command issued without a data connection", 1, "Se12");

	int file;
	if((file = open(argument, O_RDONLY)) == -1) return errno;
	sendString(connectFD, "A\n", 2);

	int i;
	while(1){
		char buffer[FILEIO] = {0};
		i = read(file, buffer, FILEIO-1);
		if(i == -1) return errno;
		if(i == 0) break;
		sendString(*dataFD, buffer, i);
	}

	if(close(file) == -1) error(strerror(errno), 1, "Se18");
	if(close(*dataFD) == -1) error(strerror(errno), 1, "Se19");
	*dataFD = -1;
	return 0;
}


int PCMD(char* argument, int* dataFD, int connectFD){
	if(*dataFD == -1) error("P command issued without a data connection", 1, "Se20");

	int file;
	if((file = open(argument, O_WRONLY | O_EXCL | O_CREAT, 00744)) == -1){
		if(close(*dataFD) == -1) error(strerror(errno), 1, "Se14");
		*dataFD = -1;
		return errno;
	}
	sendString(connectFD, "A\n", 2);

	char buffer[FILEIO];
	int bytes;
	while((bytes = read(*dataFD, buffer, FILEIO)) != 0){
		if(bytes == -1) error(strerror(errno), 1, "Se15");
		sendString(file, buffer, bytes);
	}

	if(close(file) == -1) error(strerror(errno), 1, "Se16");
	if(close(*dataFD) == -1) error(strerror(errno), 1, "Se17");
	*dataFD = -1;
	return 0;
}